@@ -33,8 +33,8 @@ public:
     bool showingDesktop = false;
     QList<PlasmaWindow *> windows;
     PlasmaWindow *activeWindow = nullptr;
-    QVector<quint32> stackingOrder;
-    QVector<QByteArray> stackingOrderUuids;
+    QList<quint32> stackingOrder;
+    QList<QByteArray> stackingOrderUuids;
 
     void setup(org_kde_plasma_window_management *wm);
 
@@ -46,8 +46,8 @@ private:
     static void stackingOrderUuidsCallback(void *data, org_kde_plasma_window_management *org_kde_plasma_window_management, const char *uuids);
     void setShowDesktop(bool set);
     void windowCreated(org_kde_plasma_window *id, quint32 internalId, const char *uuid);
-    void setStackingOrder(const QVector<quint32> &ids);
-    void setStackingOrder(const QVector<QByteArray> &uuids);
+    void setStackingOrder(const QList<quint32> &ids);
+    void setStackingOrder(const QList<QByteArray> &uuids);
 
     static struct org_kde_plasma_window_management_listener s_listener;
     PlasmaWindowManagement *q;
@@ -91,6 +91,7 @@ public:
     QStringList plasmaActivities;
     QRect geometry;
     quint32 pid = 0;
+    QString resourceName;
     QString applicationMenuServiceName;
     QString applicationMenuObjectPath;
 
@@ -98,6 +99,7 @@ private:
     static void titleChangedCallback(void *data, org_kde_plasma_window *window, const char *title);
     static void appIdChangedCallback(void *data, org_kde_plasma_window *window, const char *app_id);
     static void pidChangedCallback(void *data, org_kde_plasma_window *window, uint32_t pid);
+    static void resourceNameChangedCallback(void *data, org_kde_plasma_window *window, const char *resourceName);
     static void stateChangedCallback(void *data, org_kde_plasma_window *window, uint32_t state);
     static void virtualDesktopChangedCallback(void *data, org_kde_plasma_window *window, int32_t number);
     static void themedIconNameChangedCallback(void *data, org_kde_plasma_window *window, const char *name);
@@ -237,20 +239,23 @@ void PlasmaWindowManagement::Private::windowCreated(org_kde_plasma_window *id, q
     PlasmaWindow *window = new PlasmaWindow(q, id, internalId, uuid);
     window->d->wm = q;
     windows << window;
-    QObject::connect(window, &QObject::destroyed, q, [this, window] {
+
+    const auto windowRemoved = [this, window] {
         windows.removeAll(window);
         if (activeWindow == window) {
             activeWindow = nullptr;
             Q_EMIT q->activeWindowChanged();
         }
-    });
-    QObject::connect(window, &PlasmaWindow::unmapped, q, [this, window] {
-        if (activeWindow == window) {
-            activeWindow = nullptr;
-            Q_EMIT q->activeWindowChanged();
-        }
-    });
+    };
+
+    QObject::connect(window, &QObject::destroyed, q, windowRemoved);
+    // unmapped is emitted earlier than QObject::destroyed. We want to update windows earlier to ensure other slot will see the up to date value of
+    // PlasmaWindowManagement::windows().
+    QObject::connect(window, &PlasmaWindow::unmapped, q, windowRemoved);
     QObject::connect(window, &PlasmaWindow::activeChanged, q, [this, window] {
+        if (window->d->unmapped) {
+            return;
+        }
         if (window->isActive()) {
             if (activeWindow == window) {
                 return;
@@ -268,12 +273,9 @@ void PlasmaWindowManagement::Private::windowCreated(org_kde_plasma_window *id, q
 
 void PlasmaWindowManagement::Private::stackingOrderCallback(void *data, org_kde_plasma_window_management *interface, wl_array *ids)
 {
-    auto wm = reinterpret_cast<PlasmaWindowManagement::Private *>(data);
-    Q_ASSERT(wm->wm == interface);
-    QVector<quint32> destination;
-    destination.resize(ids->size / sizeof(uint32_t));
-    memcpy(destination.data(), ids->data, ids->size);
-    wm->setStackingOrder(destination);
+    // This is no-op since setStackingOrder(const QList<quint32> &ids) is deprecated since 5.73,
+    // but we can't remove this method because it's needed in
+    // PlasmaWindowManagement::Private::s_listener struct
 }
 
 void PlasmaWindowManagement::Private::stackingOrderUuidsCallback(void *data, org_kde_plasma_window_management *interface, const char *uuids)
@@ -283,16 +285,7 @@ void PlasmaWindowManagement::Private::stackingOrderUuidsCallback(void *data, org
     wm->setStackingOrder(QByteArray(uuids).split(';').toVector());
 }
 
-void PlasmaWindowManagement::Private::setStackingOrder(const QVector<quint32> &ids)
-{
-    if (stackingOrder == ids) {
-        return;
-    }
-    stackingOrder = ids;
-    Q_EMIT q->stackingOrderChanged();
-}
-
-void PlasmaWindowManagement::Private::setStackingOrder(const QVector<QByteArray> &uuids)
+void PlasmaWindowManagement::Private::setStackingOrder(const QList<QByteArray> &uuids)
 {
     if (stackingOrderUuids == uuids) {
         return;
@@ -397,12 +390,7 @@ PlasmaWindowModel *PlasmaWindowManagement::createWindowModel()
     return new PlasmaWindowModel(this);
 }
 
-QVector<quint32> PlasmaWindowManagement::stackingOrder() const
-{
-    return d->stackingOrder;
-}
-
-QVector<QByteArray> PlasmaWindowManagement::stackingOrderUuids() const
+QList<QByteArray> PlasmaWindowManagement::stackingOrderUuids() const
 {
     return d->stackingOrderUuids;
 }
@@ -424,6 +412,7 @@ org_kde_plasma_window_listener PlasmaWindow::Private::s_listener = {
     appmenuChangedCallback,
     activityEnteredCallback,
     activityLeftCallback,
+    resourceNameChangedCallback,
 };
 
 void PlasmaWindow::Private::appmenuChangedCallback(void *data, org_kde_plasma_window *window, const char *service_name, const char *object_path)
@@ -522,15 +511,23 @@ void PlasmaWindow::Private::pidChangedCallback(void *data, org_kde_plasma_window
     p->pid = pid;
 }
 
-void PlasmaWindow::Private::virtualDesktopChangedCallback(void *data, org_kde_plasma_window *window, int32_t number)
+void PlasmaWindow::Private::resourceNameChangedCallback(void *data, org_kde_plasma_window *window, const char *resourceName)
 {
     Q_UNUSED(window)
     Private *p = cast(data);
-    if (p->desktop == static_cast<quint32>(number)) {
+    const QString s = QString::fromUtf8(resourceName);
+    if (s == p->resourceName) {
         return;
     }
-    p->desktop = number;
-    Q_EMIT p->q->virtualDesktopChanged();
+    p->resourceName = s;
+    Q_EMIT p->q->resourceNameChanged();
+}
+
+void PlasmaWindow::Private::virtualDesktopChangedCallback([[maybe_unused]] void *data,
+                                                          [[maybe_unused]] org_kde_plasma_window *window,
+                                                          [[maybe_unused]] int32_t number)
+{
+    // Can't remove this method as it's used in PlasmaWindow::Private::s_listener struct
 }
 
 void PlasmaWindow::Private::unmappedCallback(void *data, org_kde_plasma_window *window)
@@ -631,16 +628,14 @@ static int readData(int fd, QByteArray &data)
     int n;
     while (true) {
         n = QT_READ(fd, buf, sizeof buf);
-        if (n == -1 && (errno == EAGAIN) && ++retryCount < 1000) {
+        if (n > 0) {
+            data.append(buf, n);
+        } else if (n == -1 && (errno == EAGAIN) && ++retryCount < 1000) {
             usleep(1000);
         } else {
             break;
         }
     }
-    if (n > 0) {
-        data.append(buf, n);
-        n = readData(fd, data);
-    }
     return n;
 }
 
@@ -908,16 +903,16 @@ quint32 PlasmaWindow::pid() const
     return d->pid;
 }
 
+QString PlasmaWindow::resourceName() const
+{
+    return d->resourceName;
+}
+
 QString PlasmaWindow::title() const
 {
     return d->title;
 }
 
-quint32 PlasmaWindow::virtualDesktop() const
-{
-    return d->desktop;
-}
-
 bool PlasmaWindow::isActive() const
 {
     return d->active;
@@ -1053,11 +1048,6 @@ void PlasmaWindow::requestResize()
     org_kde_plasma_window_request_resize(d->window);
 }
 
-void PlasmaWindow::requestVirtualDesktop(quint32 desktop)
-{
-    org_kde_plasma_window_set_virtual_desktop(d->window, desktop);
-}
-
 void PlasmaWindow::requestToggleKeepAbove()
 {
     if (d->keepAbove) {
@@ -1094,6 +1084,15 @@ void PlasmaWindow::requestToggleMaximized()
     }
 }
 
+void PlasmaWindow::requestToggleFullscreen()
+{
+    if (d->fullscreen) {
+        org_kde_plasma_window_set_state(d->window, ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_FULLSCREEN, 0);
+    } else {
+        org_kde_plasma_window_set_state(d->window, ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_FULLSCREEN, ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_FULLSCREEN);
+    }
+}
+
 void PlasmaWindow::setMinimizedGeometry(Surface *panel, const QRect &geom)
 {
     org_kde_plasma_window_set_minimized_geometry(d->window, *panel, geom.x(), geom.y(), geom.width(), geom.height());
@@ -1113,11 +1112,6 @@ void PlasmaWindow::requestToggleShaded()
     }
 }
 
-quint32 PlasmaWindow::internalId() const
-{
-    return d->internalId;
-}
-
 QByteArray PlasmaWindow::uuid() const
 {
     return d->uuid;
@@ -1323,3 +1317,5 @@ PlasmaActivation::PlasmaActivation(PlasmaActivationFeedback *parent, org_kde_pla
 PlasmaActivation::~PlasmaActivation() = default;
 }
 }
+
+#include "moc_plasmawindowmanagement.cpp"
